
import sprite
import random
import pygame
import math

class Kid(object):
    _DIRECTION_RIGHT = False
    _DIRECTION_LEFT = True
    _ANIMATION_IDLE = 0
    _ANIMATION_RUN = 1
    
    _SPEED = 8
    
    def __init__(self):
        boy_surface = sprite.load_image("kid.bmp", 0xffffff)
        self.sprite = sprite.AnimatedSprite(boy_surface, 8, 2)
        
        self.facing = Kid._DIRECTION_RIGHT
        
        self.hit_rect = pygame.Rect(0, 0, 50, 50)
        self.draw_rect = pygame.Rect(0, 0, 69, 71)
        self.hit_rect.bottom = 496
        self.hit_rect.centerx = 400
        self.draw_rect.bottom = 500
        self.draw_rect.centerx = 400
        
        self.good_hit_counter = 0
        self.bad_hit_counter = 0
        self.is_alive = True
        
    def tick(self, move_left, move_right):
        if self.bad_hit_counter > 0:
            self._tick_bad_hit()
        elif self.good_hit_counter > 0:
            self._tick_good_hit()
        else:
            self._tick_free(move_left, move_right)
       
    def _tick_free(self, move_left, move_right):
        if move_left and move_right:
            # if you can't decide where to go you won't move at all
            move_left = False
            move_right = False
        
        if move_left:
            self.facing = Kid._DIRECTION_LEFT
            animation = Kid._ANIMATION_RUN
            direction = -1
        elif move_right:
            self.facing = Kid._DIRECTION_RIGHT
            animation = Kid._ANIMATION_RUN
            direction = 1
        else:
            animation = Kid._ANIMATION_IDLE
            direction = 0
        # next animation frame:    
        self.sprite.next_frame(animation, self.facing)
        # move sprite on screen:
        x_off = direction * Kid._SPEED
        self.hit_rect.move_ip(x_off, 0)
        self.draw_rect.move_ip(x_off, 0)
    
    def _tick_good_hit(self):
        #TODO: use dedicated good hit animation
        #TODO: restart animation when hit twice (?)
        self.sprite.next_frame(Kid._ANIMATION_IDLE, self.facing)
        self.good_hit_counter -= 1
    
    def _tick_bad_hit(self):
        #TODO: use dedicated bad hit animation
        #TODO: restart animation when hit twice (?)
        self.sprite.next_frame(Kid._ANIMATION_IDLE, self.facing)
        self.bad_hit_counter -= 1
        if (self.bad_hit_counter < 1): self.is_alive = False
        
    def draw(self, surface):
        self.sprite.draw(surface, self.draw_rect)
        # debug output:
        # pygame.draw.rect(surface, 0x00ff00, self.hit_rect, 1)
        
    def good_hit(self):
        self.good_hit_counter = 8
    
    def bad_hit(self):
        self.bad_hit_counter = 8
        

class ThingsEmitter(object):
    MAX_DEPTH = 500
    GOOD_HIT_INCREMENT = 10
    GOOD_MISSED_DECREMENT = 1
    
    def __init__(self, target):
        self.good_list = []
        self.bad_list = []
        self.good_image = sprite.load_image("good.bmp", 0xffffff)
        self.bad_image = sprite.load_image("bad.bmp", 0xffffff)
        self.target = target
        self.random_things = ("good", "good", "bad",
                              "", "", "", "", "",
                              "", "", "", "", "",
                              "", "", "", "", "")
        self.random_velocities = range(9, 13)
        self.elapsed_ticks = 0
        self.score = 0
        
        
    def tick(self):
        self.good_list = [thing for thing in self.good_list if self.process_good_object(thing)]
        self.bad_list = [thing for thing in self.bad_list if self.process_bad_object(thing)]
        
        self.elapsed_ticks += 1
        self.level_number = math.log10(self.elapsed_ticks)
        
        self.random_new_things()
        
        
    def random_new_things(self):
        for i in range(int(self.level_number)):
            where = self.get_random_start_point()
            what = random.choice(self.random_things)
            how_fast = random.choice(self.random_velocities)
            if what == "good":
                self.good_list.append(FallingThing(self.good_image, where, how_fast))
            elif what == "bad":
                self.bad_list.append(FallingThing(self.bad_image, where, how_fast))
                
        
    
    def process_good_object(self, good_object):
        good_object.fall()
        if good_object.collides_with(self.target):
            self.target.good_hit()
            self.score += self.GOOD_HIT_INCREMENT
            return False
        elif good_object.has_reached(self.MAX_DEPTH):
            self.score -= self.GOOD_MISSED_DECREMENT
            return False
        else:
            return True
    
    def process_bad_object(self, bad_object):
        bad_object.fall()
        if bad_object.collides_with(self.target):
            self.target.bad_hit()
            return False
        elif bad_object.has_reached(self.MAX_DEPTH):
            return False
        else:
            return True
    
    def get_random_start_point(self):
        return (int(random.uniform(0,800)), 0)
    
    def draw(self, surface):
        [goodThing.draw(surface) for goodThing in self.good_list]
        [badThing.draw(surface) for badThing in self.bad_list]

   
class FallingThing(object):
    def __init__(self, surface, center_point, velocity):
        self.source_surface = surface
        self.hit_rect = pygame.Rect(0, 0, 12, 32)
        self.draw_rect = pygame.Rect(0, 0, 32, 32)
        self.hit_rect.center = center_point
        self.draw_rect.center = center_point
        self.velocity = velocity
        
    def fall(self):
        self.draw_rect.move_ip(0, self.velocity)
        self.hit_rect.move_ip(0, self.velocity)
        
    def has_reached(self, y_max):
        return self.hit_rect.bottom >= y_max
        
    def draw(self, surface):
        surface.blit(self.source_surface, self.draw_rect)
        # debug output
        # pygame.draw.rect(surface, 0x0000ff, self.hit_rect, 1)
        
    def collides_with(self, other):
        return self.hit_rect.colliderect(other.hit_rect)
        